import os
import sys
import itertools
import datetime
import psycopg2
import psycopg2.extras
import json

from log import log, log_yield, _logger
from conn_adapter import ConnAdapter
import setup

__all__ = ['Queue']


class LoggingCursor(psycopg2.extensions.cursor):
    def execute(self, sql, args=None):
        log(at='exec_sql', sql=self.mogrify(sql, args))
        try:
            super(LoggingCursor, self).execute(sql, args)
        except Exception, exc:
            log(error=repr(exc))
            raise

class LoggingRealDictCursor(LoggingCursor, psycopg2.extras.RealDictCursor):
    pass

# The queue class maps a queue abstraction onto a database table.
class Queue(object):
    def __init__(self, name, top_bound=None):
        if top_bound is None:
            top_bound = os.environ.get('QC_TOP_BOUND', 9)
        self.name, self.top_bound = name, top_bound

    @property
    def conn_adapter(self):
        if not hasattr(self, '_adapter'):
            self._adapter = ConnAdapter()
        return self._adapter

    @conn_adapter.setter
    def conn_adapter(self, conn_adapter):
        self._adapter = conn_adapter

    # enqueue(m,a) inserts a row into the jobs table and trigger a notification
    # The job's queue is represented by a name column in the row.
    # There is a trigger on the table which will send a NOTIFY event
    # on a channel which corresponds to the name of the queue.
    # The method argument is a string encoded ruby expression. The expression
    # will be separated by a `.` character and then `eval`d.
    # Examples of the method argument include: `puts`, `Kernel.puts`,
    # `MyObject.new.puts`.
    # The args argument will be encoded as JSON and stored as a JSON datatype
    # in the row. (If the version of PG does not support JSON,
    # then the args will be stored as text.
    # The args are stored as a collection and then splatted inside the worker.
    # Examples of args include: `'hello world'`, `['hello world']`,
    # `'hello', 'world'`.
    def enqueue(self, method, args):
        with log_yield(measure='queue.enqueue'):
            args = json.dumps(args)
            with self.conn_adapter.connection\
                    .cursor(cursor_factory=LoggingCursor) as curs:
                curs.execute(
                    'INSERT INTO "queue_classic_jobs" (q_name, method, args) '
                    'VALUES (%s, %s, %s) RETURNING id',
                    [self.name, method, args])
                return curs.fetchone()[0]

    def lock(self, top_bound=None):
        with log_yield(measure='queue.lock'):
            if top_bound is None:
                top_bound = self.top_bound
            with self.conn_adapter.connection\
                    .cursor(cursor_factory=LoggingRealDictCursor) as curs:
                curs.execute(
                    "SELECT * FROM lock_head(%s, %s)", [self.name, top_bound])
                if not curs.rowcount:
                    return None
                job = curs.fetchone()
            # NOTE: JSON in args is parsed automatically
            #       timestamptz columns are converted automatically to datetime
            if job['created_at']:
                now = datetime.datetime.now(job['created_at'].tzinfo)
                ttl = now - job['created_at']
                _logger.info("measure#qc.time-to-lock=%sms source=%s"
                             % (int(ttl.microseconds / 1000), self.name))
            return job

    def unlock(self, id):
        with log_yield(measure='queue.unlock'):
            return self.conn_adapter.execute(
                'UPDATE "queue_classic_jobs" '
                'SET locked_at = NULL WHERE id = %s', [id])

    def delete(self, id):
        with log_yield(measure='queue.delete'):
            return self.conn_adapter.execute(
                'DELETE FROM "queue_classic_jobs" WHERE id = %s', [id])

    def delete_all(self):
        with log_yield(measure='queue.delete_all'):
            return self.conn_adapter.execute(
                'DELETE FROM "queue_classic_jobs" WHERE q_name = %s',
                [self.name])

    def count(self):
        with log_yield(measure='queue.count'):
            with self.conn_adapter.connection\
                    .cursor(cursor_factory=psycopg2.extensions.cursor) as curs:
                curs.execute('SELECT COUNT(*) FROM "queue_classic_jobs" '
                             'WHERE q_name = %s', [self.name])
                return curs.fetchone()[0]



import os

from conn_adapter import ConnAdapter


Root = os.path.dirname(__file__)
SqlFunctions = os.path.join(Root, 'sql', 'ddl.sql')
CreateTable = os.path.join(Root, 'sql', 'create_table.sql')
DropSqlFunctions = os.path.join(Root, 'sql', 'drop_ddl.sql')

def create(connection=None):
    conn_adapter = ConnAdapter(connection)
    conn_adapter.execute(open(CreateTable).read())
    conn_adapter.execute(open(SqlFunctions).read())
    if connection is None:
        conn_adapter.disconnect()

def drop(connection=None):
    conn_adapter = ConnAdapter(connection)
    conn_adapter.execute(open(DropSqlFunctions).read())
    if connection is None:
        conn_adapter.disconnect()



import os
import sys
import datetime
import importlib
import psycopg2

from log import log, log_yield, _logger
from conn_adapter import ConnAdapter
from queue import Queue

__all__ = ['Worker']


# A Worker object can process jobs from one or many queues.
class Worker(object):

    # Creates a new worker but does not start the worker. See Worker#start.
    # This method takes a single hash argument. The following keys are read:
    # fork_worker:: Worker forks each job execution.
    # wait_interval:: Time to wait between failed lock attempts
    # connection:: PGConn object.
    # q_name:: Name of a single queue to process.
    # q_names:: Names of queues to process. Will process left to right.
    # top_bound:: Offset to the head of the queue. 1 == strict FIFO.
    def __init__(self, fork_worker=None, wait_interval=None, connection=None,
                 q_name=None, q_names=None, top_bound=None):
        if fork_worker is None:
            fork_worker = bool(os.environ.get('QC_FORK_WORKER', ''))
        if wait_interval is None:
            wait_interval = int(os.environ.get('QC_LISTEN_TIME', '5'))
        self.fork_worker = fork_worker
        self.wait_interval = wait_interval
        self.conn_adapter = ConnAdapter(connection)
        if q_name is None:
            q_name = os.environ.get('QUEUE', 'default')
        if q_names is None:
            q_names = os.environ.get('QUEUES', '')
            if not q_names:
                q_names = []
            else:
                q_names = q_names.split(',')
        self.queues = self.__setup_queues(
            self.conn_adapter, q_name, q_names, top_bound)
        self.running = True
        log(at="worker_initialized")

    # Commences the working of jobs.
    # start() spins on @running -which is initialized as true.
    # This method is the primary entry point to starting the worker.
    # The canonical example of starting a worker is as follows:
    # QC::Worker.new.start
    def start(self):
        while self.running:
            if self.fork_worker:
                self.fork_and_work()
            else:
                self.work()

    # Signals the worker to stop taking new work.
    # This method has no immediate effect. However, there are
    # two loops in the worker (one in #start and another in #lock_job)
    # which check the @running variable to determine if further progress
    # is desirable. In the case that @running is false, the aforementioned
    # methods will short circuit and cause the blocking call to #start
    # to unblock.
    def stop(self):
        self.running = False

    # Calls Worker#work but after the current process is forked.
    # The parent process will wait on the child process to exit.
    def fork_and_work(self):
        cpid = os.fork()
        if cpid == 0:
            try:
                self.setup_child()
                self.work()
            except:
                # prevent going up in the stack
                os._exit(1)
            else:
                # prevent going up in the stack
                os._exit(0)
        else:
            log(at="fork", pid=str(cpid))
            os.waitpid(cpid, 0)

    # Blocks on locking a job, and once a job is locked,
    # it will process the job.
    def work(self):
        queue, job = self.lock_job()
        if queue and job:
            log(at="work", job=str(job['id']))
            self.process(queue, job)

    # Attempt to lock a job in the queue's table.
    # If a job can be locked, this method returns an array with
    # 2 elements. The first element is the queue from which the job was locked
    # and the second is a hash representation of the job.
    # If a job is returned, its locked_at column has been set in the
    # job's row. It is the caller's responsibility to delete the job row
    # from the table when the job is complete.
    def lock_job(self):
        log(at="lock_job")
        job = None
        while self.running:
            for queue in self.queues:
                job = queue.lock()
                if job:
                    return (queue, job)
            self.conn_adapter.wait(self.wait_interval,
                *[queue.name for queue in self.queues])

    # A job is processed by evaluating the target code.
    # if the job is evaluated with no exceptions
    # then it is deleted from the queue.
    # If the job has raised an exception the responsibility of what
    # to do with the job is delegated to Worker#handle_failure.
    # If the job is not finished and an INT signal is trapped,
    # this method will unlock the job in the queue.
    def process(self, queue, job):
        start = datetime.datetime.now()
        finished = False
        try:
            self.call(job)
            queue.delete(job['id'])
            finished = True
        except Exception, e:
            self.handle_failure(job, e)
            finished = True
        finally:
            if not finished:
                queue.unlock(job['id'])
            ttp = datetime.datetime.now() - start
            _logger.info("measure#qc.time-to-process=%s source=%s"
                         % (int(ttp.microseconds / 1000), queue.name))

    # Each job includes a method column. We will use ruby's eval
    # to grab the ruby object from memory. We send the method to
    # the object and pass the args.
    def call(self, job):
        args = job['args']
        receiver_str, _, message = job['method'].rpartition('.')
        if receiver_str:
            module = importlib.import_module(receiver_str, self.__module__)
        else:
            import __main__
            module = __main__
        getattr(module, message)(*args)

    # This method will be called when an exception
    # is raised during the execution of the job.
    def handle_failure(self, job, e):
        _logger.error("count#qc.job-error=1 job=%s error=%s"
                      % (repr(job), repr(e)))

    # This method should be overriden if
    # your worker is forking and you need to
    # re-establish database connections
    def setup_child(self):
        log(at="setup_child")

    def log(self, data):
        log(data)

    def __setup_queues(self, conn_adapter, queue, queues, top_bound):
        names = (queues if len(queues) > 0 else [queue])
        queues = [Queue(name, top_bound) for name in names]
        for queue in queues:
            queue.conn_adapter = conn_adapter
        return queues



import datetime
import logging
from contextlib import contextmanager


_logger = logging.getLogger('pueuey')

def log(**data):
    data = dict(data, lib='pueuey')
    _logger.debug(" ".join([(k + '=' + v) for k, v in data.items()]))

@contextmanager
def log_yield(**data):
    try:
        t0 = datetime.datetime.now()
        yield
    except Exception, e:
        log(**dict(data, at='error', error=repr(e)))
        raise
    else:
        t = int((datetime.datetime.now() - t0).microseconds / 1000)
        log(**dict(data, elapsed=(str(t) + 'ms')))



import os
import re
import psycopg2
import select
import urlparse

from log import log

__all__ = ['ConnAdapter']


class ConnAdapter(object):
    def __str__(self):
        return "<%s object at 0x%x; dsn: %s, closed: %d>" \
            % (self.__class__.__name__, id(self),
               repr(self.connection.dsn), self.connection.closed)

    def __init__(self, connection=None):
        self.connection = (
            self.__establish_new()
            if connection is None
            else self.__validate(connection)
        )
        self.connection.set_isolation_level(
            psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)

    def disconnect(self):
        try:
            return self.connection.close()
        except Exception, e:
            log(at='disconnect', error=repr(e))
            raise

    def execute(self, *args):
        log(at='exec_sql', sql=repr(args[0]))
        try:
            with self.connection.cursor() as curs:
                curs.execute(*args)
        except Exception, e:
            log(error=repr(e))
            raise

    def wait(self, time, *channels):
        curs = self.connection.cursor()
        listen_cmds = ['LISTEN "%s"' % c for c in channels]
        curs.execute(';'.join(listen_cmds))
        channel = self.__wait_for_notify(time)
        unlisten_cmds = ['UNLISTEN "%s"' % c for c in channels]
        curs.execute(';'.join(unlisten_cmds))
        self.__drain_notify()
        return channel

    def __drain_notify(self):
        notifies = list(self.connection.notifies)
        if notifies:
            log(at='drain_notifications')
            del self.connection.notifies[:]
        return notifies

    def __wait_for_notify(self, t):
        if any(select.select([self.connection], [], [], t)):
            self.connection.poll()
            if self.connection.notifies:
                return self.connection.notifies.pop()
        return None

    def __validate(self, connection):
        assert isinstance(connection, psycopg2._psycopg.connection), \
            "connection must be an instance of " \
            "psycopg2._psycopg.connection, but was " + repr(type(connection))
        return connection

    def __establish_new(self):
        log(at='establish_conn')
        try:
            conn = psycopg2.connect(**self.__normalize_db_url(self.__db_url()))
        except Exception, e:
            log(error=repr(e))
            raise
        with conn.cursor() as curs:
            curs.execute("SET application_name = %s",
                [os.environ.get('QC_APP_NAME', 'queue_classic')])
        return conn

    def __normalize_db_url(self, url):
        host = url.hostname
        if host:
            host = re.sub(r"%2F", "/", host, flags=re.I)

        return {
            'host': host, # host or percent-encoded socket path
            'port': url.port,
            'database': re.sub(r"/", "", url.path),
            'username' : url.username,
            'password' : url.password,
        }

    def __db_url(self):
        url = os.environ.get('QC_DATABASE_URL', os.environ.get('DATABASE_URL'))
        if not url:
            raise ValueError("missing QC_DATABASE_URL or DATABASE_URL")
        return urlparse.urlparse(url)



from setuptools import setup, find_packages

setup(
    name = "pueuey",
    version = "1.0",
    author = "Cecile Tonglet",
    author_email = "cecile.tonglet@gmail.com",
    description = ("queue_classic ported to Python 2: Simple, efficient "
                   "worker queue for Ruby & PostgreSQL."),
    license = "MIT",
    keywords = "postgres queue worker ruby",
    url = "https://github.com/cecton/pueuey",
    packages = find_packages(),
    install_requires = ['psycopg2'],
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Topic :: Database",
        "Topic :: System :: Distributed Computing",
        "License :: OSI Approved :: MIT License",
    ],
)
